@@ -47,7 +47,8 @@
 	 pkix_is_fixed_dh_cert/1,
 	 pkix_is_issuer/2,
 	 pkix_issuer_id/2,
-	 pkix_normalize_name/1,
+	 pkix_subject_id/1,
+         pkix_normalize_name/1,
 	 pkix_path_validation/3,
 	 pkix_verify_hostname/2, pkix_verify_hostname/3,
          pkix_verify_hostname_match_fun/1,
@@ -62,12 +63,15 @@
 	 pkix_crl_issuer/1,
 	 short_name_hash/1,
          pkix_test_data/1,
-         pkix_test_root_cert/2
+         pkix_test_root_cert/2,
+         pkix_ocsp_validate/5,
+         ocsp_responder_id/1,
+         ocsp_extensions/1
 	]).
 
 -export_type([public_key/0, private_key/0, pem_entry/0,
 	      pki_asn1_type/0, asn1_type/0, ssh_file/0, der_encoded/0,
-              key_params/0, digest_type/0, issuer_name/0, oid/0]).
+              key_params/0, digest_type/0, issuer_name/0, cert_id/0, oid/0]).
 
 -type public_key()           ::  rsa_public_key() | rsa_pss_public_key() | dsa_public_key() | ec_public_key() | ed_public_key() .
 -type private_key()          ::  rsa_private_key() | rsa_pss_private_key() | dsa_private_key() | ec_private_key() | ed_private_key() .
@@ -123,8 +127,7 @@
 -type oid()                  :: tuple().
 -type chain_type()           :: server_chain | client_chain.
 
--type issuer_id()            :: {SerialNr::integer(), issuer_name()} .
-
+-type cert_id()              :: {SerialNr::integer(), issuer_name()} .
 -type issuer_name()          :: {rdnSequence,[#'AttributeTypeAndValue'{}]} .
 
 
@@ -905,7 +908,7 @@ pkix_is_fixed_dh_cert(Cert) when is_binary(Cert) ->
 
 %%--------------------------------------------------------------------
 -spec pkix_issuer_id(Cert, IssuedBy) ->
-			    {ok, issuer_id()} | {error, Reason}
+			    {ok, ID::cert_id()} | {error, Reason}
                                 when Cert::der_encoded()| #'OTPCertificate'{},
                                      IssuedBy :: self | other,
                                      Reason :: term() .
@@ -919,6 +922,19 @@ pkix_issuer_id(Cert, Signed) when is_binary(Cert) ->
     OtpCert = pkix_decode_cert(Cert, otp),
     pkix_issuer_id(OtpCert, Signed).
 
+%%--------------------------------------------------------------------
+-spec pkix_subject_id(Cert) -> ID
+              when Cert::der_encoded()| #'OTPCertificate'{},
+                   ID::cert_id() .
+
+%% Description: Returns the subject id.
+%%--------------------------------------------------------------------
+pkix_subject_id(#'OTPCertificate'{} = OtpCert) ->
+    pubkey_cert:subject_id(OtpCert);
+pkix_subject_id(Cert) when is_binary(Cert) -> 
+    OtpCert = pkix_decode_cert(Cert, otp),
+    pkix_subject_id(OtpCert).
+
 %%--------------------------------------------------------------------
 -spec pkix_crl_issuer(CRL| #'CertificateList'{}) -> 
                              Issuer when CRL :: der_encoded(),
@@ -945,7 +961,7 @@ pkix_normalize_name(Issuer) ->
 
 %%-------------------------------------------------------------------- 
 -spec pkix_path_validation(Cert::binary()| #'OTPCertificate'{} | atom(),
-			   CertChain :: [binary()] ,
+			   CertChain :: [binary() | #'OTPCertificate'{}] ,
 			   Options :: [{atom(),term()}]) ->
 				  {ok, {PublicKeyInfo :: term(), 
 					PolicyTree :: term()}} |
@@ -962,8 +978,8 @@ pkix_path_validation(PathErr, [Cert | Chain], Options0) when is_atom(PathErr)->
 	    Options = proplists:delete(verify_fun, Options0),
 	    pkix_path_validation(Otpcert, Chain, [{verify_fun,
 						   {VerifyFun, Userstate}}| Options]);
-	{fail, _} ->
-	    {error, Reason}
+	{fail, UserReason} ->
+	    {error, UserReason}
     catch
 	_:_ ->
 	    {error, Reason}
@@ -976,10 +992,19 @@ pkix_path_validation(TrustedCert, CertChain, Options)
 pkix_path_validation(#'OTPCertificate'{} = TrustedCert, CertChain, Options)
   when is_list(CertChain), is_list(Options) ->
     MaxPathDefault = length(CertChain),
-    ValidationState = pubkey_cert:init_validation_state(TrustedCert, 
-							MaxPathDefault, 
-							Options),
-    path_validation(CertChain, ValidationState).
+    {VerifyFun, Userstat0} =
+	proplists:get_value(verify_fun, Options, ?DEFAULT_VERIFYFUN),
+    try pubkey_cert:validate_time(TrustedCert, Userstat0, VerifyFun) of
+        Userstate1 -> 
+            ValidationState = pubkey_cert:init_validation_state(TrustedCert, 
+                                                                MaxPathDefault, 
+                                                                [{verify_fun, {VerifyFun, Userstate1}} | 
+                                                                 proplists:delete(verify_fun, Options)]),
+            path_validation(CertChain, ValidationState)
+    catch
+        throw:{bad_cert, cert_expired} = Reason ->
+            {error, Reason}    
+    end.
 
 %--------------------------------------------------------------------
 -spec pkix_crls_validate(OTPcertificate, DPandCRLs, Options) ->
@@ -1261,6 +1286,47 @@ pkix_test_data(#{} = Chain) ->
 
 pkix_test_root_cert(Name, Opts) ->
     pubkey_cert:root_cert(Name, Opts).
+      
+%%--------------------------------------------------------------------
+-spec pkix_ocsp_validate(Cert, IssuerCert, OcspRespDer, 
+                         ResponderCerts, NonceExt) -> valid | {bad_cert, Reason}
+              when Cert::der_encoded() | #'OTPCertificate'{},
+                   IssuerCert::der_encoded() | #'OTPCertificate'{}, 
+                   OcspRespDer::der_encoded(),
+                   ResponderCerts::[der_encoded()],
+                   NonceExt::undefined | binary(),
+                   Reason::term().
+
+%% Description: Validate OCSP staple response
+%%--------------------------------------------------------------------
+pkix_ocsp_validate(DerCert, IssuerCert, OcspRespDer, ResponderCerts, NonceExt) when is_binary(DerCert) ->
+    pkix_ocsp_validate(pkix_decode_cert(DerCert, otp),  IssuerCert, OcspRespDer, ResponderCerts, NonceExt);
+pkix_ocsp_validate(Cert, DerIssuerCert, OcspRespDer, ResponderCerts, NonceExt) when is_binary(DerIssuerCert) ->
+    pkix_ocsp_validate(Cert, pkix_decode_cert(DerIssuerCert, otp), OcspRespDer, ResponderCerts, NonceExt);
+pkix_ocsp_validate(Cert, IssuerCert, OcspRespDer, ResponderCerts, NonceExt) ->
+    case  ocsp_responses(OcspRespDer, ResponderCerts, NonceExt) of
+        {ok, Responses} ->
+            ocsp_status(Cert, IssuerCert, Responses);
+        {error, Reason} ->
+            {bad_cert, {revocation_status_undetermined, Reason}}
+    end.
+
+%%--------------------------------------------------------------------
+-spec ocsp_extensions(undefined | binary()) -> list().
+%% Description: Get OCSP stapling extensions for request
+%%--------------------------------------------------------------------
+ocsp_extensions(Nonce) ->
+    [Extn || Extn <- [pubkey_ocsp:get_nonce_extn(Nonce),
+                      pubkey_ocsp:get_acceptable_response_types_extn()],
+             erlang:is_record(Extn, 'Extension')].
+
+%%--------------------------------------------------------------------
+-spec ocsp_responder_id(#'Certificate'{}) -> binary().
+%%
+%% Description: Get the OCSP responder ID der
+%%--------------------------------------------------------------------
+ocsp_responder_id(Cert) ->
+    pubkey_ocsp:get_ocsp_responder_id(Cert).
 
 %%--------------------------------------------------------------------
 %%% Internal functions
@@ -1387,16 +1453,16 @@ path_validation([Cert | _] = Path,
     end.
 
 validate(Cert, #path_validation_state{working_issuer_name = Issuer,
-					 working_public_key = Key,
-					 working_public_key_parameters = 
-					 KeyParams, 
-					 permitted_subtrees = Permit,
-					 excluded_subtrees = Exclude,
-					 last_cert = Last,
-					 user_state = UserState0,
-					 verify_fun = VerifyFun} =
+                                      working_public_key = Key,
+                                      working_public_key_parameters = 
+                                          KeyParams, 
+                                      permitted_subtrees = Permit,
+                                      excluded_subtrees = Exclude,
+                                      last_cert = Last,
+                                      user_state = UserState0,
+                                      verify_fun = VerifyFun} =
 	     ValidationState0) ->
-
+    
     OtpCert = otp_cert(Cert),
 
     {ValidationState1, UserState1} =
@@ -1758,8 +1824,8 @@ verify_hostname_match_default0(_, _) ->
 
 
 verify_hostname_match_wildcard(FQDN, Name) ->
-    [F1|Fs] = string:tokens(FQDN, "."),
-    [N1|Ns] = string:tokens(Name, "."),
+    [F1|Fs] = string:tokens(to_lower_ascii(FQDN), "."),
+    [N1|Ns] = string:tokens(to_lower_ascii(Name), "."),
     match_wild(F1,N1) andalso Fs==Ns.
 
 
@@ -1824,3 +1890,14 @@ format_details([]) ->
 format_details(Details) ->
     Details.
   
+ocsp_status(Cert, IssuerCert, Responses) ->
+    case pubkey_ocsp:find_single_response(Cert, IssuerCert, Responses) of
+        {ok, #'SingleResponse'{certStatus = CertStatus}} ->
+            pubkey_ocsp:ocsp_status(CertStatus);
+        {error, no_matched_response = Reason} ->
+            {bad_cert, {revocation_status_undetermined, Reason}}
+    end.
+
+ocsp_responses(OCSPResponseDer, ResponderCerts, Nonce) ->
+    pubkey_ocsp:verify_ocsp_response(OCSPResponseDer, 
+                                     ResponderCerts, Nonce).
